@@ -1,7 +1,7 @@
 package Catmandu::BagIt;
 
 use strict;
-our $VERSION = '0.12';
+our $VERSION = '0.13';
 
 use Catmandu::Sane;
 use Catmandu;
@@ -12,9 +12,8 @@ use IO::File qw();
 use IO::Handle qw();
 use File::Copy;
 use List::MoreUtils qw(first_index uniq);
-use File::Path qw(remove_tree mkpath);
-use File::Slurper qw(read_lines write_text);
-use File::Temp qw(tempfile);
+use Path::Tiny;
+use Path::Iterator::Rule;
 use Catmandu::BagIt::Payload;
 use Catmandu::BagIt::Fetch;
 use POSIX qw(strftime);
@@ -22,35 +21,40 @@ use LWP::UserAgent;
 use utf8;
 use namespace::clean;
 
+# Flags indicating which operations are needed to create a valid bag
 use constant {
-    FLAG_BAGIT        => 0x001 ,
-    FLAG_BAG_INFO     => 0x002 ,
-    FLAG_FETCH        => 0x004 ,
-    FLAG_DATA         => 0x008 , 
-    FLAG_TAG_MANIFEST => 0x016 ,
-    FLAG_MANIFEST     => 0x032 ,
-    FLAG_DIRTY        => 0x064 ,
+    FLAG_BAGIT        => 0x001 , # Flag indicates updating the bagit.txt file required
+    FLAG_BAG_INFO     => 0x002 , # Flag indicates updating the bag-info.txt file required
+    FLAG_FETCH        => 0x004 , # Flag indicates updating the fetch.txt file required
+    FLAG_DATA         => 0x008 , # Flag indicating new payload data available
+    FLAG_TAG_MANIFEST => 0x016 , # Flag indicates updateing tag-manifest-manifest.txt required
+    FLAG_MANIFEST     => 0x032 , # Flag indicates updating manifest-md5.txt required
+    FLAG_DIRTY        => 0x064 , # Flag indicates payload file that hasn't been serialized
 };
 
 with 'Catmandu::Logger';
 
+# Array containing all errors when reading/writing bags
 has '_error' => (
     is       => 'rw',
     default  => sub { [] },
 );
 
+# Integer containing a combinatation of FLAG_* set for this bag
 has 'dirty' => (
     is       => 'ro',
     writer   => '_dirty',
     default  => 0,
 );
 
+# Path to a directory containing a bag
 has 'path' => (
     is       => 'ro',
     writer   => '_path',
     init_arg => undef,
 );
 
+# Version number of the bag specification
 has 'version' => (
     is       => 'ro',
     writer   => '_version',
@@ -58,6 +62,7 @@ has 'version' => (
     init_arg => undef,
 );
 
+# Encoding of all tag manifests
 has 'encoding' => (
     is       => 'ro',
     writer   => '_encoding',
@@ -65,90 +70,100 @@ has 'encoding' => (
     init_arg => undef,
 );
 
-has user_agent => (is => 'ro');
+# User agent used to fetch payloads from the Internet
+has user_agent => (is => 'lazy');
 
+# An array of a tag file names
 has '_tags' => (
     is       => 'rw',
     default  => sub { [] },
     init_arg => undef,
 );
 
+# An array of Catmandu::BagIt::Payloads
 has '_files' => (
     is       => 'rw',
     default  => sub { [] },
     init_arg => undef,
 );
 
+# An array of Catmandu::BagIt::Fetch
 has '_fetch' => (
     is       => 'rw',
     default  => sub { [] },
     init_arg => undef,
 );
 
+# A lookup hash of md5 checksums for the tag files
 has '_tag_sums' => (
     is       => 'rw',
     default  => sub { {} },
     init_arg => undef,
 );
 
+# A lookup hahs of md5 checksums for the payload files
 has '_sums' => (
     is       => 'rw',
     default  => sub { {} },
     init_arg => undef,
 );
 
+# An array of hashes of all name/value pairs in the bag-info.txt file
 has '_info' => (
     is       => 'rw',
     default  => sub { [] },
     init_arg => undef,
 );
 
-has _http_client => (
-    is => 'ro', 
-    lazy => 1, 
-    builder => '_build_http_client', 
-    init_arg => 'user_agent'
-);
-
-sub _build_http_client {
+sub _build_user_agent {
     my ($self) = @_;
     my $ua = LWP::UserAgent->new;
     $ua->agent('Catmandu-BagIt/' . $Catmandu::BagIt::VERSION);
     $ua;
 }
 
+# Settings requires when creating a new bag from scratch
 sub BUILD {
     my $self = shift;
 
     $self->log->debug("initializing bag");
 
+    # Intialize the in memory settings of the bag-info
     $self->_update_info;
+
+    # Initialize the in memory settings of the tag-manifests
     $self->_update_tag_manifest;
+
+    # Intialize the names of the basic tag files
     $self->_tags([qw(
             bagit.txt
             bag-info.txt
             manifest-md5.txt
             )]);
 
-
+    # Set this bag as dirty requiring an update of all the files
     $self->_dirty($self->dirty | FLAG_BAG_INFO | FLAG_TAG_MANIFEST | FLAG_DATA | FLAG_BAGIT);
 }
 
+# Return all the arrors as an array
 sub errors {
     my ($self) = @_;
     @{$self->_error};
 }
 
+# Return an array of tag file names
 sub list_tags {
     my ($self) = @_;
     @{$self->_tags};
 }
 
+# Return an array of all Catmandu::BagIt::Payload-s
 sub list_files {
     my ($self) = @_;
     @{$self->_files};
 }
 
+# Return a Catmandu::BagIt::Payload given a file name
 sub get_file {
     my ($self,$filename) = @_;
     die "usage: get_file(filename)" unless $filename;
@@ -159,6 +174,7 @@ sub get_file {
     return undef;
 }
 
+# Return a Catmandu::BagIt::Fetch given a file name
 sub get_fetch {
     my ($self,$filename) = @_;
     die "usage: get_fetch(filename)" unless $filename;
@@ -169,26 +185,32 @@ sub get_fetch {
     return undef;
 }
 
+# Return true when this bag is dirty
 sub is_dirty {
     my ($self) = @_;
     $self->dirty != 0;
 }
 
+# Return true when this bag is holey (and requires fetching data from the Internet
+# to be made complete)
 sub is_holey {
     my ($self) = @_;
     @{$self->_fetch} > 0;
 }
 
+# Return an array of Catmandu::BagIt::Fetch
 sub list_fetch {
     my ($self) = @_;
     @{$self->_fetch};
 }
 
+# Return an array of tag file
 sub list_tagsum {
     my ($self) = @_;
     keys %{$self->_tag_sums};
 }
 
+# Return the md5 checksum of a file
 sub get_tagsum {
     my ($self,$file) = @_;
 
@@ -197,11 +219,13 @@ sub get_tagsum {
     $self->_tag_sums->{$file};
 }
 
+# Return an array of payload files
 sub list_checksum {
     my ($self) = @_;
     keys %{$self->_sums};
 }
 
+# Return the md5 checksum of of a file name
 sub get_checksum {
     my ($self,$file) = @_;
 
@@ -210,6 +234,7 @@ sub get_checksum {
     $self->_sums->{$file};
 }
 
+# Read the content of a bag
 sub read {
     my ($class,$path) = @_;
 
@@ -242,6 +267,7 @@ sub read {
     $ok == 7 ? $self : undef;
 }
 
+# Write the content of a bag back to disk
 sub write {
     my ($self,$path,%opts) = @_;
 
@@ -250,19 +276,18 @@ sub write {
     die "usage: write(path[, overwrite => 1])" unless $path;
 
     # Check if other processes are writing or previous processes died
-    if ( 
-        (defined($self->path) && -f $self->path . "/.lock") || 
-        -f "$path/.lock"
-       ) {
-        $self->log->error($self->path . "/.lock or $path/.lock exists");
-        $self->_push_error($self->path . "/.lock or $path/.lock exists");
+    if ($self->locked($path)) {
+        $self->log->error("$path is locked");
+        $self->_push_error("$path is locked");
         return undef;
     }
 
     if (defined($self->path) && $path ne $self->path) {
+        # If the bag is copied from to a new location than all the tag files and
+        # files should be flagged as dirty and need to be overwritten
         $self->log->info("copying from old path: " . $self->path);
         $self->_dirty($self->dirty | FLAG_BAGIT | FLAG_BAG_INFO | FLAG_TAG_MANIFEST | FLAG_MANIFEST | FLAG_DATA);
-        
+
         foreach my $item ($self->list_files) {
             $item->flag($item->flag ^ FLAG_DIRTY);
         }
@@ -272,11 +297,12 @@ sub write {
         # updates are possible when overwrite => 1
     }
     elsif ($opts{overwrite} && -d $path) {
+        # Remove existing bags
         $self->log->info("removing: $path");
-        remove_tree($path);
+        path($path)->remove_tree;
     }
 
-    if (-f "$path/bagit.txt") {
+    if (-f $self->_bagit_file($path)) {
         if ($opts{overwrite}) {
             $self->log->info("overwriting: $path");
         }
@@ -288,12 +314,12 @@ sub write {
     }
     else {
         $self->log->info("creating: $path");
-        mkpath($path);
+        path($path)->mkpath;
         $self->_dirty($self->dirty | FLAG_BAGIT);
     }
 
-    unless ($self->touch("$path/.lock")) {
-        $self->log->error("failed to create $path/.lock");
+    unless ($self->touch($self->_lock_file($path))) {
+        $self->log->error("failed to lock in $path");
         return undef;
     }
 
@@ -308,11 +334,78 @@ sub write {
     $ok += $self->_write_manifest($path);
     $ok += $self->_write_tag_manifest($path);
 
+    return undef unless $ok == 6;
+
     $self->_dirty(0);
 
-    unlink("$path/.lock");
+    unlink($self->_lock_file($path));
 
-    $ok == 6;
+    $ok = 0;
+
+    # Reread the contents of the bag
+    $ok += $self->_read_version($path);
+    $ok += $self->_read_info($path);
+    $ok += $self->_read_tag_manifest($path);
+    $ok += $self->_read_manifest($path);
+    $ok += $self->_read_tags($path);
+    $ok += $self->_read_files($path);
+    $ok += $self->_read_fetch($path);
+
+    $ok == 7;
+}
+
+sub _bagit_file {
+    my ($self,$path) = @_;
+
+    File::Spec->catfile($path,'bagit.txt');
+}
+
+sub _bag_info_file {
+    my ($self,$path) = @_;
+
+    File::Spec->catfile($path,'bag-info.txt');
+}
+
+sub _package_info_file {
+    my ($self,$path) = @_;
+
+    File::Spec->catfile($path,'package-info.txt');
+}
+
+sub _manifest_md5_file {
+    my ($self,$path) = @_;
+
+    File::Spec->catfile($path,'manifest-md5.txt');
+}
+
+sub _tagmanifest_md5_file {
+    my ($self,$path) = @_;
+
+    File::Spec->catfile($path,'tagmanifest-md5.txt');
+}
+
+sub _fetch_file {
+    my ($self,$path) = @_;
+
+    File::Spec->catfile($path,'fetch.txt');
+}
+
+sub _tag_file {
+    my ($self,$path,$file) = @_;
+
+    File::Spec->catfile($path,$file);
+}
+
+sub _payload_file {
+    my ($self,$path,$file) = @_;
+
+    File::Spec->catfile($path,'data',$file);
+}
+
+sub _lock_file {
+    my ($self,$path) = @_;
+
+    File::Spec->catfile($path,'.lock');
 }
 
 sub locked {
@@ -321,7 +414,7 @@ sub locked {
 
     return undef unless defined($path);
 
-    -f "$path/.lock";
+    -f $self->_lock_file($path);
 }
 
 sub touch {
@@ -329,10 +422,8 @@ sub touch {
 
     die "usage: touch(path)"
             unless defined($path);
-    local(*F);
-    open(F,">$path") || die "failed to open $path for writing: $!";
-    print F "";
-    close(F);
+
+    path("$path")->spew("");
 
     1;
 }
@@ -340,7 +431,7 @@ sub touch {
 sub add_file {
     my ($self, $filename, $data, %opts) = @_;
 
-    die "usage: add_file(filename, data [, overwrite => 1])" 
+    die "usage: add_file(filename, data [, overwrite => 1])"
             unless defined($filename) && defined($data);
 
     $self->_error([]);
@@ -363,13 +454,18 @@ sub add_file {
         return;
     }
 
-    push @{ $self->_files } , Catmandu::BagIt::Payload->new(
-                                    filename => $filename , 
-                                    data => $data ,
-                                    flag => FLAG_DIRTY ,
-                                );
+    my $payload = Catmandu::BagIt::Payload->from_any($filename,$data);
+    $payload->flag(FLAG_DIRTY);
 
-    my $sum = $self->_md5_sum($data);
+    push @{ $self->_files }, $payload;
+
+    my $fh = $payload->open;
+
+    binmode($fh,":raw");
+
+    my $sum = $self->_md5_sum($fh);
+
+    close($fh);
 
     $self->_sums->{"$filename"} = $sum;
 
@@ -422,7 +518,7 @@ sub remove_file {
 sub add_fetch {
     my ($self, $url, $size, $filename) = @_;
 
-    die "usage add_fetch(url,size,filename)" 
+    die "usage add_fetch(url,size,filename)"
             unless defined($url) && $size =~ /^\d+$/ && defined($filename);
 
     die "illegal file name $filename"
@@ -447,17 +543,15 @@ sub add_fetch {
 sub remove_fetch {
     my ($self, $filename) = @_;
 
-    die "usage remove_fetch(filename)" unless defined($filename);  
+    die "usage remove_fetch(filename)" unless defined($filename);
 
     $self->log->info("removing fetch for $filename");
 
     my (@old) = grep { $_->filename ne $filename} @{$self->_fetch};
 
     $self->_fetch(\@old);
-
     $self->_update_info;
     $self->_update_tag_manifest;
-
     $self->_dirty($self->dirty | FLAG_FETCH | FLAG_TAG_MANIFEST);
 
     1;
@@ -469,15 +563,15 @@ sub mirror_fetch {
     die "usage mirror_fetch(<Catmandu::BagIt::Fetch>)"
             unless defined($fetch) && ref($fetch) && ref($fetch) =~ /^Catmandu::BagIt::Fetch/;
 
-    my ($tmp_fh, $tmp_filename) = tempfile();
+    my $tmp_filename = Path::Tiny->tempfile;
 
     my $url       = $fetch->url;
     my $filename  = $fetch->filename;
-    my $path      = $self->path;  
+    my $path      = $self->path;
 
     $self->log->info("mirroring $url -> $tmp_filename...");
 
-    my $response = $self->_http_client->mirror($url,$tmp_filename);
+    my $response = $self->user_agent->mirror($url,$tmp_filename);
 
     if ($response->is_success) {
         $self->log->info("mirror is a success");
@@ -495,8 +589,8 @@ sub mirror_fetch {
 sub add_info {
     my ($self,$name,$values) = @_;
 
-    die "usage add_info(name,values)" 
-            unless defined($name) && defined($values);  
+    die "usage add_info(name,values)"
+            unless defined($name) && defined($values);
 
     if ($name =~ /^(Bag-Size|Bagging-Date|Payload-Oxum)$/) {
         for my $part (@{$self->_info}) {
@@ -530,20 +624,20 @@ sub add_info {
 sub remove_info {
     my ($self,$name) = @_;
 
-    die "usage remove_info(name)" 
-            unless defined($name); 
+    die "usage remove_info(name)"
+            unless defined($name);
 
     if ($name =~ /^(Bag-Size|Bagging-Date|Payload-Oxum)$/) {
         $self->log->error("removing info $name - is read-only");
         return undef;
-    } 
+    }
 
     $self->log->info("removing info $name");
 
     my (@old) = grep { $_->[0] ne $name } @{$self->_info};
 
     $self->_info(\@old);
-    
+
     $self->_update_tag_manifest;
 
     $self->_dirty($self->dirty | FLAG_BAG_INFO | FLAG_TAG_MANIFEST);
@@ -636,8 +730,8 @@ sub complete {
 
     foreach my $file (@missing) {
         unless (grep { $_->filename =~ /^$file$/ } $self->list_fetch) {
-            $self->log->error("file $file doesn' exist in bag and fetch.txt");
-            $self->_push_error("file $file doesn' exist in bag and fetch.txt");
+            $self->log->error("file $file doesn't exist in bag and fetch.txt");
+            $self->_push_error("file $file doesn't exist in bag and fetch.txt");
         }
     }
 
@@ -663,18 +757,20 @@ sub valid {
         }
 
         my $md5 = $tag == 0 ? $self->get_checksum($file) : $self->get_tagsum($file);
-        my $fh  = $tag == 0 ? new IO::File "$path/data/$file", "r" : new IO::File "$path/$file" , "r";
+        my $fh  = $tag == 0 ?
+                    new IO::File $self->_payload_file($path,$file), "r" :
+                    new IO::File $self->_tag_file($path,$file) , "r";
 
         unless ($fh) {
             $self->log->error("can't read $file");
             return (0,"can't read $file");
         }
 
-        binmode($fh);
+        binmode($fh,':raw');
 
         my $md5_check = $self->_md5_sum($fh);
 
-        undef $fh;
+        close($fh);
 
         unless ($md5 eq $md5_check) {
             $self->log->error("$file checksum fails $md5 <> $md5_check");
@@ -726,14 +822,10 @@ sub _size {
 
     my $total = 0;
 
-    foreach my $item ($self->list_files) {
-        my $size;
-        if ($item->is_io && $item->data->can('stat')) {
-            $size = [ $item->data->stat ]->[7];
-        }
-        else {
-            $size = length($item->data);
-        }
+    foreach my $file ($self->list_files) {
+        my $fh   = $file->open;
+        my $size = [ $fh->stat ]->[7];
+        $fh->close;
         $total += $size;
     }
 
@@ -796,11 +888,11 @@ sub _read_fetch {
 
     $self->_fetch([]);
 
-    return 1 unless -f "$path/fetch.txt";
+    return 1 unless -f $self->_fetch_file($path);
 
     $self->log->debug("reading fetch.txt");
 
-    foreach my $line (read_lines("$path/fetch.txt",'UTF-8')) {
+    foreach my $line (path($self->_fetch_file($path))->lines_utf8) {
         $line =~ s/\r\n$/\n/g;
         chomp($line);
 
@@ -819,13 +911,13 @@ sub _read_tag_manifest {
 
     $self->_tag_sums({});
 
-    if (! -f "$path/tagmanifest-md5.txt") {
+    if (! -f $self->_tagmanifest_md5_file($path)) {
         return 1;
     }
 
     $self->log->debug("reading tagmanifest-md5.txt");
 
-    foreach my $line (read_lines("$path/tagmanifest-md5.txt",'UTF-8')) {
+    foreach my $line (path($self->_tagmanifest_md5_file($path))->lines_utf8) {
        $line =~ s/\r\n$/\n/g;
         chomp($line);
         my ($sum,$file) = split(/\s+/,$line,2);
@@ -842,12 +934,12 @@ sub _read_manifest {
 
     $self->_sums({});
 
-    if (! -f "$path/manifest-md5.txt") {
-        $self->_push_error("$path/manifest-md5.txt bestaat niet");
+    if (! -f $self->_manifest_md5_file($path)) {
+        $self->_push_error("no manifest-md5.txt in $path");
         return 0;
     }
 
-    foreach my $line (read_lines("$path/manifest-md5.txt",'UTF-8')) {
+    foreach my $line (path($self->_manifest_md5_file($path))->lines_utf8) {
         $line =~ s/\r\n$/\n/g;
         chomp($line);
         my ($sum,$file) = split(/\s+/,$line,2);
@@ -865,20 +957,18 @@ sub _read_tags {
 
     $self->_tags([]);
 
-    local(*F);
-    
-    open(F,"find $path -maxdepth 1 -type f |") || die "can't find tag-files";
+    my $rule = Path::Iterator::Rule->new;
+    $rule->max_depth(1);
+    $rule->file;
+    my $iter = $rule->iter($path);
 
-    while(<F>) {
-        chomp($_);
-        $_ =~ s/^$path.//;
+    while(my $file = $iter->()) {
+        $file =~ s/^$path.//;
 
-        next if $_ =~ /^tagmanifest-\w+.txt$/;
+        next if $file =~ /^tagmanifest-\w+.txt$/;
 
-        push @{ $self->_tags } , $_;
+        push @{ $self->_tags } , $file;
     }
-    
-    close(F);
 
     1;
 }
@@ -896,20 +986,17 @@ sub _read_files {
         return 1;
     }
 
-    local(*F);
-    open(F,"find $path/data -type f |") || die "payload directory doesn't contain files";
+    my $rule = Path::Iterator::Rule->new;
+    $rule->file;
+    my $iter = $rule->iter("$path/data");
 
-    while(my $file = <F>) {
-        chomp($file);
+    while(my $file = $iter->()) {
         my $filename = $file;
         $filename =~ s/^$path\/data\///;
-        my $data = IO::File->new($file);
-
-        push @{ $self->_files } , Catmandu::BagIt::Payload->new(filename => $filename, data => $data);
+        my $payload = Catmandu::BagIt::Payload->new(filename => $filename, path => $file);
+        push @{ $self->_files } , $payload;
     }
 
-    close(F);
-
     1;
 }
 
@@ -920,15 +1007,17 @@ sub _read_info {
 
     $self->_info([]);
 
-    my $info_file = -f "$path/bag-info.txt" ? "$path/bag-info.txt" :  "$path/package-info.txt";
+    my $info_file = -f $self->_bag_info_file($path) ?
+                        $self->_bag_info_file($path) :
+                        $self->_package_info_file($path);
 
     if (! -f $info_file) {
-        $self->log->error("$path/package-info.txt or $path/bag-info.txt doesn't exist");
-        $self->_push_error("$path/package-info.txt or $path/bag-info.txt doesn't exist");
+        $self->log->error("no package-info.txt or bag-info.txt in $path");
+        $self->_push_error("no package-info.txt or bag-info.txt in $path");
         return 0;
     }
 
-    foreach my $line (read_lines($info_file, 'UTF-8')) {
+    foreach my $line (path($info_file)->lines_utf8) {
         $line =~ s/\r\n$/\n/g;
         chomp($line);
 
@@ -951,14 +1040,14 @@ sub _read_version {
 
     $self->log->debug("reading the version file");
 
-    if (! -f "$path/bagit.txt" ) {
-        $self->log->error("$path/bagit.txt doesn't exist");
-        $self->_push_error("$path/bagit.txt doesn't exist");
+    if (! -f $self->_bagit_file($path) ) {
+        $self->log->error("no bagit.txt in $path");
+        $self->_push_error("no bagit.txt in $path");
         return 0;
     }
 
-    foreach my $line (read_lines("$path/bagit.txt",'UTF-8')) {
-    $line =~ s/\r\n$/\n/g;
+    foreach my $line (path($self->_bagit_file($path))->lines_utf8) {
+        $line =~ s/\r\n$/\n/g;
         chomp($line);
         my ($n,$v) = split(/\s*:\s*/,$line,2);
 
@@ -980,16 +1069,7 @@ sub _write_bagit {
 
     $self->log->info("writing the version file");
 
-    local (*F);
-    unless (open(F,">:utf8" , "$path/bagit.txt")) {
-        $self->log->error("can't create $path/bagit.txt: $!");
-        $self->_push_error("can't create $path/bagit.txt: $!");
-        return;
-    }
-
-    printf F $self->_bagit_as_string;
-
-    close (F);
+    path($self->_bagit_file($path))->spew_utf8($self->_bagit_as_string);
 
     $self->_dirty($self->dirty ^ FLAG_BAGIT);
 
@@ -1015,17 +1095,7 @@ sub _write_info {
 
     $self->log->info("writing the tag info file");
 
-    local(*F);
-
-    unless (open(F,">:utf8", "$path/bag-info.txt")) {
-        $self->log->error("can't create $path/bag-info.txt: $!");
-        $self->_push_error("can't create $path/bag-info.txt: $!");
-        return;
-    }
-
-    print F $self->_baginfo_as_string;
-
-    close(F);
+    path($self->_bag_info_file($path))->spew_utf8($self->_baginfo_as_string);
 
     $self->_dirty($self->dirty ^ FLAG_BAG_INFO);
 
@@ -1053,13 +1123,16 @@ sub _baginfo_as_string {
     $str;
 }
 
+# Write BagIt data payloads to disk
 sub _write_data {
     my ($self,$path) = @_;
 
+    # Return immediately when no files need to be written
     return 1 unless $self->dirty & FLAG_DATA;
 
     $self->log->info("writing the data files");
 
+    # Create a data/ directory for payloads
     unless (-d "$path/data") {
         unless (mkdir "$path/data") {
             $self->log->error("can't create payload directory $path/data: $!");
@@ -1068,58 +1141,51 @@ sub _write_data {
         }
     }
 
+    # Create a list of all files written to the payload directory
+    # Compare this list later with files found in the payload directory
+    # This difference are the files that can be deleted
     my @all_names_in_bag = ();
 
     foreach my $item ($self->list_files) {
         my $filename = 'data/' . $item->{filename};
         push @all_names_in_bag , $filename;
 
+        # Only process files that are dirty
         next unless $item->flag & FLAG_DIRTY;
 
+        # Check for deep directories that need to be stored
         my $dir  = $filename; $dir =~ s/\/[^\/]+$//;
 
         $self->log->info("serializing $filename");
 
-        mkpath("$path/$dir") unless -d "$path/$dir";
+        path("$path/$dir")->mkpath unless -d "$path/$dir";
 
-        if ($item->is_io) {
-            $item->fh->seek(0,0) if $item->data->can('seek');
-            eval {
-                copy($item->fh, "$path/$filename");
-            };
-            if ($@) {
-                if ($@ =~ /are identical/) {
-                    $self->log->error("attempy to copy identical files");
-                }
-            }
-            # Close the old handle
-            $item->fh->close();
-            # Reopen the file at the new position
-            $item->{data} = IO::File->new("$path/$filename");
-            $item->flag($item->flag ^ FLAG_DIRTY);
+        my $old_path = $item->path;
+        my $new_path = "$path/$filename";
+
+        if ($item->is_new) {
+            File::Copy::move($old_path,$new_path);
         }
         else {
-            write_text("$path/$filename", $item->data);
-            $item->flag($item->flag ^ FLAG_DIRTY);
+            File::Copy::copy($old_path,$new_path);
         }
+
+        $item->flag($item->flag ^ FLAG_DIRTY);
     }
 
-    # Check deleted files
-    local(*F);
-    
-    if (open(F,"find $path/data -type f |")) {
-        while(my $file = <F>) {
-            chomp($file);
-            
-            my $filename = $file;
-            $filename =~ s/^$path\///;
- 
-            unless (grep {$filename eq $_} @all_names_in_bag) {
-                $self->log->info("deleting $path/$filename");
-                unlink "$path/$filename";
-            }
+    # Check deleted files. Delete all files not in the @all_names_in_bag list
+    my $rule = Path::Iterator::Rule->new;
+    $rule->file;
+    my $iter = $rule->iter("$path/data");
+
+    while(my $file = $iter->()) {
+        my $filename = $file;
+        $filename =~ s/^$path\///;
+
+        unless (grep {$filename eq $_} @all_names_in_bag) {
+            $self->log->info("deleting $path/$filename");
+            unlink "$path/$filename";
         }
-        close(F);
     }
 
     $self->_dirty($self->dirty ^ FLAG_DATA);
@@ -1136,29 +1202,19 @@ sub _write_fetch {
 
     unless (defined($fetch_str) && length($fetch_str)) {
         $self->log->info("removing fetch.txt");
-        unlink "$path/fetch.txt" if -f "$path/fetch.txt";
+        unlink $self->_fetch_file($path) if -f $self->_fetch_file($path);
         return 1;
     }
 
     $self->log->info("writing the fetch file");
 
     if ($self->_fetch == 0) {
-        unlink "$path/fetch.txt" if -r "$path/fetch.txt";
+        unlink $self->_fetch_file($path) if -r $self->_fetch_file($path);
         $self->_dirty($self->dirty ^ FLAG_FETCH);
         return 1;
     }
 
-    local(*F);
-
-    unless (open(F,">:utf8", "$path/fetch.txt")) {
-        $self->log->error("can't create $path/fetch.txt: $!");
-        $self->_push_error("can't create $path/fetch.txt: $!");
-        return;
-    }
-
-    print F $fetch_str;
-
-    close (F);
+    path($self->_fetch_file($path))->spew_utf8($fetch_str);
 
     $self->_dirty($self->dirty ^ FLAG_FETCH);
 
@@ -1184,17 +1240,7 @@ sub _write_manifest {
 
     $self->log->info("writing the manifest file");
 
-    local (*F);
-
-    unless (open(F,">:utf8", "$path/manifest-md5.txt")) {
-        $self->log->error("can't create $path/manifest-md5.txt: $!");
-        $self->_push_error("can't create $path/manifest-md5.txt: $!");
-        return;
-    }
-
-    print F $self->_manifest_as_string;
-
-    close(F);
+    path($self->_manifest_md5_file($path))->spew_utf8($self->_manifest_as_string);
 
     $self->_dirty($self->dirty ^ FLAG_MANIFEST);
 
@@ -1210,7 +1256,7 @@ sub _manifest_as_string {
     my $str = '';
 
     foreach my $file ($self->list_checksum) {
-        next unless -f "$path/data/$file";
+        next unless -f $self->_payload_file($path,$file);
         my $md5 = $self->get_checksum($file);
         $str .= "$md5 data/$file\n";
     }
@@ -1226,19 +1272,9 @@ sub _write_tag_manifest {
     # The tag manifest can be dirty when writing new files
     $self->_update_tag_manifest;
 
-    $self->log->info("writing the tag manifest file");
+    $self->log->info("writing the tagmanifest file");
 
-    local (*F);
-
-    unless (open(F,">:utf8", "$path/tag-manifest-md5.txt")) {
-        $self->log->error("can't create $path/manifest-md5.txt: $!");
-        $self->_push_error("can't create $path/manifest-md5.txt: $!");
-        return;
-    }
-
-    print F $self->_tag_manifest_as_string;
-
-    close(F);
+    path($self->_tagmanifest_md5_file($path))->spew_utf8($self->_tag_manifest_as_string);
 
     $self->_dirty($self->dirty ^ FLAG_MANIFEST);
 
@@ -1252,7 +1288,7 @@ sub _tag_manifest_as_string {
 
     foreach my $file ($self->list_tagsum) {
         my $md5  = $self->get_tagsum($file);
-        print F "$md5 $file\n";
+        $str .= "$md5 $file\n";
     }
 
     $str;
@@ -1264,17 +1300,20 @@ sub _md5_sum {
     my $ctx = Digest::MD5->new;
 
     if (!defined $data) {
-    return $ctx->add(Encode::encode_utf8(''))->hexdigest;
+        return $ctx->add(Encode::encode_utf8(''))->hexdigest;
     }
     elsif (! ref $data) {
         return $ctx->add(Encode::encode_utf8($data))->hexdigest;
     }
-    elsif (ref $data eq 'SCALAR') {
+    elsif (ref($data) eq 'SCALAR') {
         return $ctx->add(Encode::encode_utf8($$data))->hexdigest;
     }
-    else {
+    elsif (ref($data) =~ /^IO/) {
         return $ctx->addfile($data)->hexdigest;
     }
+    else {
+        die "unknown data type: `" . ref($data) . "`";
+    }
 }
 
 sub _is_legal_file_name {
@@ -1285,17 +1324,6 @@ sub _is_legal_file_name {
     return 1;
 }
 
-sub DESTROY {
-    my ($self) = @_;
-
-    # Closing open file handles
-    foreach my $item ($self->list_files) {
-        if ($item->is_io && $item->fh->opened) {
-            $item->fh->close;
-        }
-    }
-}
-
 1;
 
 __END__
@@ -1336,38 +1364,40 @@ Catmandu::BagIt - Low level Catmandu interface to the BagIt packages.
     printf "tags:\n";
     for my $tag ($bagit->list_info_tags) {
         my @values = $bagit->get_info($tag);
-        printf " $tag: %s\n" , join(", ",@values); 
+        printf " $tag: %s\n" , join(", ",@values);
     }
 
     printf "tag-sums:\n";
     for my $file ($bagit->list_tagsum) {
         my $sum = $bagit->get_tagsum($file);
-        printf " $file: %s\n" , $sum; 
+        printf " $file: %s\n" , $sum;
     }
 
     # Read the file listing as found in the manifest file
     printf "file-sums:\n";
     for my $file ($bagit->list_checksum) {
         my $sum = $bagit->get_checksum($file);
-        printf " $file: %s\n" , $sum; 
+        printf " $file: %s\n" , $sum;
     }
 
     # Read the real listing of files as found on the disk
     printf "files:\n";
     for my $file ($bagit->list_files) {
-        my $stat = [$file->data->stat];
+        my $stat = [$file->path];
         printf " name: %s\n", $file->filename;
         printf " size: %s\n", $stat->[7];
         printf " last-mod: %s\n", scalar(localtime($stat->[9]));
     }
 
     my $file = $bagit->get_file("mydata.txt");
-    my $fh   = $file->fh;
+    my $fh   = $file->open;
 
     while (<$fh>) {
        ....
     }
 
+    close($fh);
+
     print "dirty?\n" if $bagit->is_dirty;
 
     if ($bagit->complete) {
@@ -1456,7 +1486,7 @@ Return the encoding of the BagIt.
 
 =head2 size()
 
-Return a human readble string of the expected size of the BagIt (adding the actual sizes found on disk plus 
+Return a human readble string of the expected size of the BagIt (adding the actual sizes found on disk plus
 the files that need to be fetched from the network).
 
 =head2 payload_oxum()
@@ -1473,7 +1503,7 @@ Return true when the BagIt contains a non emtpy fetch configuration.
 
 =head2 is_error()
 
-Return an ARRAY of errors when checking complete, valid and write. 
+Return an ARRAY of errors when checking complete, valid and write.
 
 =head2 complete()
 
@@ -1499,7 +1529,7 @@ Remove an info $tag.
 
 =head2 get_info($tag, [$delim])
 
-Return an ARRAY of values found for the $tag name. Or, in scalar context, return a string of 
+Return an ARRAY of values found for the $tag name. Or, in scalar context, return a string of
 all values optionally delimeted by $delim.
 
 =head2 list_tagsum()
@@ -1530,6 +1560,8 @@ Get a Catmandu::BagIt::Payload object for the file $filename.
 
 =head2 add_file($filename, IO::File->new(...))
 
+=head2 add_file($filaname, sub { my $io = shift; .... })
+
 Add a new file to the BagIt.
 
 =head2 remove_file($filename)
