@@ -95,7 +95,7 @@ use IO::File;
 use LWP::Simple;
 use Moo;
 
-our $VERSION = '0.12';
+our $VERSION = '0.13';
 
 with 'Catmandu::Exporter';
 
@@ -135,7 +135,7 @@ sub add {
 
             # For now using a simplistic mirror operation
             my $fname = $tmp->filename;
-            my $response = $bagit->_http_client->mirror($url,$fname);
+            my $response = $bagit->user_agent->mirror($url,$fname);
 
             unless ($response->is_success) {
                 Catmandu::Error->throw("failed to mirror $url to $fname : " . $response->status_line);
